﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text;
using TftpServer.Utils;
using Utils;

namespace TftpServer
{
    public class TftpPacket
    {
        public TftpOpCode OpCode { get; }

        public string FileName { get; }
        public string Mode { get; }

        public short BlockNumber { get; }
        public byte[] DataBuffer { get; }
        public int DataLength { get; }

        public TftpErrorNumber ErrorNumber { get; }
        public string ErrorMessage { get; }

        public int BlockSize
        {
            get => Options.MapValue("blksize", int.Parse, () => 512);
            set => Options["blksize"] = value.ToString();
        }
        public int? TransferSize
        {
            get => Options.MapValue("tsize", s => int.Parse(s), () => (int?)null);
            set => Options["tsize"] = value.ToString();
        }

        public Dictionary<string, string> Options { get; set; } = new Dictionary<string, string>();

        public TftpPacket(byte[] buffer, int length)
        {
            using var stream = new MemoryStream(buffer, 0, length, true, true);
            using var reader = new BinaryReader(stream);

            OpCode = (TftpOpCode)IPAddress.NetworkToHostOrder(reader.ReadInt16());

            switch (OpCode)
            {
                default: throw new NotImplementedException($"OpCode: {OpCode} (0x{(short)OpCode:X4})");

                case TftpOpCode.Read:
                    FileName = stream.ReadNullTerminatedString();
                    Mode = stream.ReadNullTerminatedString();
                    break;

                case TftpOpCode.Ack:
                    BlockNumber = IPAddress.NetworkToHostOrder(reader.ReadInt16());
                    break;

                case TftpOpCode.Error:
                    ErrorNumber = (TftpErrorNumber)IPAddress.NetworkToHostOrder(reader.ReadInt16());
                    ErrorMessage = stream.ReadNullTerminatedString();
                    break;
            }

            while (stream.Position < length)
            {
                var key = stream.ReadNullTerminatedString();
                if (stream.Position >= length) continue;
                var val = stream.ReadNullTerminatedString();
                Options[key] = val;
            }
        }

        public TftpPacket(string errorMessage, TftpErrorNumber errorNumber = TftpErrorNumber.SeeMessage) : this(TftpOpCode.Error)
        {
            ErrorMessage = errorMessage;
            ErrorNumber = errorNumber;
        }

        public TftpPacket(short blockNumber, byte[] dataBuffer, int dataLength) : this(TftpOpCode.Data)
        {
            BlockNumber = blockNumber;
            DataBuffer = dataBuffer;
            DataLength = dataLength;
        }

        public TftpPacket(TftpOpCode opCode)
        {
            OpCode = opCode;
        }

        public int WriteToBuffer(byte[] buffer)
        {
            using var stream = new MemoryStream(buffer);
            using var writer = new BinaryWriter(stream);

            writer.Write(IPAddress.HostToNetworkOrder((short)OpCode));

            switch (OpCode)
            {
                default: throw new NotImplementedException($"OpCode: {OpCode} (0x{(short)OpCode:X4})");

                case TftpOpCode.OptionAck:
                    break;

                case TftpOpCode.Data:
                    writer.Write(IPAddress.HostToNetworkOrder(BlockNumber));
                    if (0 == DataLength) writer.Write(DataBuffer);
                    else writer.Write(DataBuffer, 0, DataLength);
                    break;

                case TftpOpCode.Error:

                    writer.Write(IPAddress.HostToNetworkOrder((short)ErrorNumber));

                    if (null != ErrorMessage)
                    {
                        var len = Encoding.ASCII.GetBytes(ErrorMessage, 0, ErrorMessage.Length, buffer, (int)stream.Position);
                        stream.Position += len;
                    }
                    writer.Write((byte)0);
                    break;
            }

            foreach (var option in Options)
            {
                stream.WriteNullTerminatedString(option.Key);
                stream.WriteNullTerminatedString(option.Value);
            }

            return (int)stream.Position;
        }
    }
}